/*
 *  Copyright 2003-2014  Mikhail Kryshen <mikhail@kryshen.net>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.games1729.shariks;

import java.awt.*;
import java.awt.image.*;
import java.util.*;

class Renderer {
    private final Game game;
    private final Skin skin;
    private final Component target;

    private volatile String message = null;
    private boolean messageUpdated = false;
    private long messageClearTime = -1;

    private Image background = null;
    private Image foreground = null;
    private BufferedImage messageBuffer = null;

    Renderer(Game game) {
	this.game = game;
	this.skin = game.skin;
	this.target = game.component;
    }

    void setMessage(String text) {
	synchronized (this) {
	    this.message = text;
	    messageUpdated = true;
	    messageClearTime = -1;
	}	
	game.update();
    }
    
    void setMessage(String text, long period) {
	synchronized (this) {
	    this.message = text;
	    messageUpdated = true;
	    messageClearTime = System.currentTimeMillis() + period;
	}
	game.update();
    }

    public void render(Graphics2D g) {
	int width, height;

	if (game.field != null) {
	    width = (int)(game.field.width * game.field.getScale());
	    height = (int)(game.field.height * game.field.getScale());
 	} else {
	    width = game.getWidth();
	    height = game.getHeight();
	}

	if (skin.getBackground() != null) {
	    background = scale(skin.getBackground(), background,
			       width, height, null);
	    g.drawImage(background, 0, 0, null);
	} else {
	    // FIXME: use background color from skin.
	    ((Graphics2D)g).setBackground(Color.GRAY);
	    g.clearRect(0, 0, width, height);
	}

	if (game.field != null) {
	    game.field.render(g);
	}

	if (skin.getForeground() != null) {
	    foreground = scale(skin.getForeground(), foreground,
			       width, height, null);
	    g.drawImage(foreground, 0, 0, null);
	}

        if (message != null) {
            synchronized (this) {
                if (messageClearTime > 0
                        && messageClearTime < System.currentTimeMillis()) {

                    message = null;
                    messageClearTime = -1;
                }

                if (message != null) {
                    if (messageUpdated) {
                        if (messageBuffer == null
                                || messageBuffer.getWidth() != width
                                || messageBuffer.getHeight() != height) {

                            messageBuffer = game.component.getGraphicsConfiguration().
                                    createCompatibleImage(width, height, Transparency.TRANSLUCENT);
                        }

                        Graphics2D mg = messageBuffer.createGraphics();

                        mg.setComposite(AlphaComposite.Clear);
                        mg.fillRect(0, 0, width, height);
                        mg.setComposite(AlphaComposite.SrcOver);

                        mg.addRenderingHints(new RenderingHints(RenderingHints.KEY_TEXT_ANTIALIASING,
                                RenderingHints.VALUE_TEXT_ANTIALIAS_ON));

                        // TODO: get font and color from skin.
                        mg.setColor(Color.white);
                        mg.setFont(new Font("SansSerif", Font.BOLD, 24));
                        drawStringInRect(message, 0, 0, width, height, 2, 0, 2, mg);

                        mg.dispose();
                    }

                    g.drawImage(messageBuffer, 0, 0, null);
                }
            }
        }
    }

    Image scale(Image image, Image cached, int width, int height,
		ImageFilter filter) {
	if (cached != null &&
	    cached.getWidth(null) == width &&
	    cached.getHeight(null) == height) {
	    
	    return cached;
	}

	synchronized (skin) {
	    MediaTracker tracker = new MediaTracker(target);
	    if (filter != null) {
		image = target.createImage
		    (new FilteredImageSource(image.getSource(), filter));
	    }
	    image = image.getScaledInstance(width, height, Image.SCALE_SMOOTH);
	    tracker.addImage(image, 0);
	    try {
		tracker.waitForID(0);
	    } catch (InterruptedException e) {
		System.err.println(e);
	    }
            return image;
            
//            GraphicsConfiguration gc = target.getGraphicsConfiguration();
//            BufferedImage scaled = gc.createCompatibleImage(width, height, 
//                    Transparency.TRANSLUCENT);
//            Graphics2D g = scaled.createGraphics();
//            
//            float sx = (float) width / image.getWidth(null);
//            float sy = (float) height / image.getHeight(null);
//            
//            System.err.println(sx + "; " + sy);
//            
//            g.drawImage(image, AffineTransform.getScaleInstance(sx, sy), null);
//            g.dispose();
//            
//            return scaled;
	}
    }

    void drawStringInRect(String s, int x, int y, int w, int h, 
			       int align, int gap, int shadow, Graphics2D g) {
	final int BLUR_WIDTH = 4;

	x += gap;
	w = w - gap - gap;
	
	FontMetrics fm = g.getFontMetrics();
	StringTokenizer st = new StringTokenizer(s, "\n");

	int sh = fm.getHeight();
	int sy = y + Math.round((h - sh * st.countTokens()) / 2F) + fm.getAscent();

	while (st.hasMoreTokens()) {
	    String line = st.nextToken();
	    int sx;
	    int sw = fm.stringWidth(line);

	    switch (align) {
	    case 1:          // left
		sx = x;
	    	break;
	    case 3:          // right
		sx = x + w - sw;
		break;
	    default:         // center
		sx = x + (w - sw) / 2;
	    }
	   
	    if (shadow != 0) {
		BufferedImage si = target.getGraphicsConfiguration()
		    .createCompatibleImage(sw + BLUR_WIDTH, 
					   sh + BLUR_WIDTH,
					   Transparency.TRANSLUCENT);

		Graphics2D sig = si.createGraphics();
		sig.setRenderingHints(g.getRenderingHints());

		sig.setFont(g.getFont());
		sig.setColor(Color.black);
		sig.drawString(line, BLUR_WIDTH / 2, fm.getAscent() + BLUR_WIDTH / 2);
		sig.dispose();

		Kernel kernel = new Kernel(3, 3, new float[] {
		    1f/9f, 1f/9f, 1f/9f,
		    1f/9f, 1f/9f, 1f/9f,
		    1f/9f, 1f/9f, 1f/9f});
		BufferedImageOp op = new ConvolveOp(kernel);

		g.drawImage(si, op, sx + shadow - BLUR_WIDTH / 2, 
			    sy + shadow - BLUR_WIDTH / 2 - fm.getAscent());
	    }

	    g.drawString(line, sx, sy);
	    sy += sh;
	}       
    }
}
